/*
Author: CAI Yu
Email: icaiyu0618@gmail.com

This plugin is aimed for upload the files in multi directories for the Embulk.
To use the plugin read the ReadMe.md carefully.

This plugin will parse the directories in the multi_dir parameter in your config.yml
The embulk will load the files in each directory. And each directory will be treated as a task.
The embulk will run those tasks in multi threads.

This plugin should be used with the WendlinPlugin.

*/

package org.embulk.input.filename;

import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileTime;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;

//import com.google.common.base.Optional;
import org.slf4j.Logger;


import org.embulk.config.Config;
import org.embulk.config.ConfigDefault;
import org.embulk.config.ConfigDiff;
import org.embulk.config.ConfigSource;
import org.embulk.config.Task;
import org.embulk.config.TaskReport;
import org.embulk.config.TaskSource;
import org.embulk.config.ConfigInject;
import org.embulk.spi.PageBuilder;
import org.embulk.spi.Exec;
import org.embulk.spi.InputPlugin;
import org.embulk.spi.PageOutput;
import org.embulk.spi.Schema;
import org.embulk.spi.SchemaConfig;
import org.embulk.spi.BufferAllocator;
import org.embulk.spi.ColumnConfig;
import org.embulk.spi.SchemaConfig;

import static org.embulk.spi.type.Types.STRING;


public class FilenameInputPlugin
        implements InputPlugin
{
    public interface PluginTask
            extends Task
    {
        @Config("multi_dir")
        @ConfigDefault("[]")
        ArrayList<String> getMultiDir();
         
        @Config("multi_tag")
        @ConfigDefault("[]")
        ArrayList<String> getMultiTag();
        
        @Config("lastPaths")
        @ConfigDefault("[]")
        ArrayList<String> getLastPaths();
        
        @Config("chunk_size")
        @ConfigDefault("10485760")
        int getChunkSize();
        
        @Config("load_order")
        @ConfigDefault("\"\"")
        String getLoadOrder();


        // Not implements yet this Configuration will decide whether upload the symlinks files.
        @Config("follow_symlinks")
        @ConfigDefault("false")
        boolean getFollowSymlinks();
        

        ArrayList<ArrayList<String>> getFiles();
        void setFiles(ArrayList<ArrayList<String>> allFiles);

        @ConfigInject
        BufferAllocator getBufferAllocator();
       
       
    }
    
    // initialize a log. Very useful tool to display the important information.
    private final Logger log = Exec.getLogger(getClass());
    
    // Get the CURRENT directory of the embulk runing.
    private final static Path CURRENT_DIR = Paths.get(".").normalize();
    
    // This varibale record the tags for each directory
    private static ArrayList<String> tagList;
    
    
    // This varibale decide the chunkSize of the upload data.
    private static int chunkSize;
    
    
    // You need to learn the workflow of the embulk before understand how this transaction work.
    @Override
    public ConfigDiff transaction(ConfigSource config,
            InputPlugin.Control control)
    {
        // load the task from the TaskSource. from this varibale, we could read the configuration in the config.yml
        PluginTask task = config.loadConfig(PluginTask.class);
        
        // Get the data chunk size
        chunkSize = task.getChunkSize();
        
        // Read the directories list from the task.
        ArrayList<String> dirList = task.getMultiDir();
        
        // Read the LastPath list from the task
        ArrayList<String> lastPaths = task.getLastPaths();
        
        // We create a big Array to contains all directories, and each directories will contain many files.
        ArrayList<ArrayList<String>> allFiles = new ArrayList<ArrayList<String>>();
        
        // Read the tags list from the task.
        tagList = task.getMultiTag();
        
        // If the dirList have no directory, we throw an RuntimeException.
        if ( dirList.size() != 0 ){
            log.info ("The list of the directories: " + dirList );
            
            // If the Number of tags is less than the directories, we say that the default tag for the directory is ""
            while (tagList.size() < dirList.size()){
                tagList.add("");
            }
            
            // if the number of lastPaths is less than the directory, we say the default lastPaths for the directory is ""
            while (lastPaths.size() < dirList.size()){
                lastPaths.add("");
            }
        } else {
            throw new RuntimeException("The multi_dir should contain at least 1 directory.");
        }
        
        
        // Now to time to read all files from each directory.
        for (int i =0; i < dirList.size();i++ ){
            String dir = dirList.get(i);
            String lastPath = lastPaths.get(i);
            
            // We have to sort the files before we set them to the tasks.
            // Here we get the parameter about the order
            
            Comparator comparator;
            String order = task.getLoadOrder();
            switch (order) {
                case "ASCEND_MODIFIED": comparator = AscendModifiedSorter.getComparator();
                break;
                case "DESCEND_MODIFIED": comparator = DescendModifiedSorter.getComparator();
                break;
                case "ASCEND_CREATION": comparator = AscendCreationSorter.getComparator();
                break;
                case "DESCEND_CREATION": comparator = DescendModifiedSorter.getComparator();
                break;
                default: comparator = new Comparator<String>();
                break;
            }
            
            
            
            // This method return the files in a directory,however the files in the varibale files is in random order. we have to sort them next
            ArrayList<String> files = listFiles(task,Paths.get(dir).normalize(),lastPath,order);
            
            
            
            // Sort the files for each directory
            Collections.sort(files,comparator);
        
            
            // End of sort
            
            log.info("The files is " + files);
            
            // add the files to the big Array.
            allFiles.add(files);
            
        }
        
        // each directory will be treated as a task. the taskCount is the size of the bigArray.
        int taskCount = allFiles.size();
        log.info("taskCount of the input plugin is: " + taskCount);
        // We set the allFiles to the tasks.files.
        task.setFiles(allFiles);
        
        // Here we add we columns for the Schema.
        ArrayList<ColumnConfig> columns = new ArrayList<ColumnConfig>();
        
        // Here we add two columns to the columns.
        // if you want, we can read the columnName from the seed.yml: final String columnName = task.getColumnName();
        columns.add(new ColumnConfig("payload", STRING, config));
        columns.add(new ColumnConfig("tag", STRING, config));
        
        // Create a Schema for the Page.
        Schema schema = new SchemaConfig(columns).toSchema();

        return resume(task.dump(), schema, taskCount, control);
    }

    @Override
    public ConfigDiff resume(TaskSource taskSource,
            Schema schema, int taskCount,
            InputPlugin.Control control)
    {
        control.run(taskSource, schema, taskCount);
        return Exec.newConfigDiff();
    }

    @Override
    public void cleanup(TaskSource taskSource,
            Schema schema, int taskCount,
            List<TaskReport> successTaskReports)
    {
    }


    // This function will be run in every task. As we say that we distribute each directory as task.
    // in this method. we have to load all files in the directory.
    @Override
    public TaskReport run(TaskSource taskSource,
            Schema schema, int taskIndex,
            PageOutput output)
    {
        PluginTask task = taskSource.loadTask(PluginTask.class);
        
        ArrayList<String> files = task.getFiles().get(taskIndex);
        
        // Check how many files in this directory on the run.
        log.info("The files in the run:" + files);
        
        // For each file, we create some pages.
        // If the size of the file is less than chunkSize, we build just one page for this file.
        // If the size of the file is more than chunkSize, we build more than one page for this file.
        for (String file : files)
        {
            try
            {
                int nRead;
                byte[] data = new byte[chunkSize];
                FileInputStream dataIn  = new  FileInputStream(file);
                ByteArrayOutputStream buffer = new ByteArrayOutputStream();
                
                // Read the data and build the page.
                while ((nRead = dataIn.read(data, 0, data.length)) != -1) {
                    buffer.write(data, 0, nRead);
                    try (PageBuilder pageBuilder = new PageBuilder(Exec.getBufferAllocator(), schema, output))
                    {
                        pageBuilder.setString(0,buffer.toString());//Base64.encodeBase64String(buffer.toByteArray()));
                        pageBuilder.setString(1, tagList.get(taskIndex) + new File(file).getCanonicalPath() );
                        pageBuilder.addRecord();
                        buffer.flush();
                        pageBuilder.finish();
                    }
                }
            } catch (IOException ex){
                ex.printStackTrace();
            }
        }

        TaskReport taskReport = Exec.newTaskReport();
        taskReport.set("columns", schema.size());
        return taskReport;
    }

    @Override
    public ConfigDiff guess(ConfigSource config)
    {
        return Exec.newConfigDiff();
    }
    
    // This method is for walk through the directory and record the files in the directory. It will compare the filename with the lastPath
    // In we want to upload the files in ALPHABETICAL order. than the filename "smaller than" the lastPath will be abandonned.
    // Be careful, that since we have alternative for the order. You should be careful what "smaller than" means!
    public ArrayList<String> listFiles(PluginTask task,Path pathPrefix,final String lastPath,final String order)
    {
        //Path pathPrefix = Paths.get(task.getPathPrefix()).normalize();
        final Path directory;
        final String fileNamePrefix;
        if (Files.isDirectory(pathPrefix)) {
            directory = pathPrefix;
            fileNamePrefix = "";
        } else {
            fileNamePrefix = pathPrefix.getFileName().toString();
            Path d = pathPrefix.getParent();
            directory = (d == null ? CURRENT_DIR : d);
        }

        final ArrayList<String> filesArray = new ArrayList<String>();
        try {
            log.info("Listing local files at directory '{}' filtering filename by prefix '{}'", directory.equals(CURRENT_DIR) ? "." : directory.toString(), fileNamePrefix);
            Files.walkFileTree(directory, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path path, BasicFileAttributes attrs)
                {
                    if (path.equals(directory)) {
                        return FileVisitResult.CONTINUE;
                    } else if (lastPath != null && path.toString().compareTo(lastPath.substring(0, path.toString().length())) < 0) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            return FileVisitResult.CONTINUE;
                        } else {
                            return FileVisitResult.SKIP_SUBTREE;
                        }
                    }
                }


                @Override
                public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
                {
                    try
                    {
                        if ( !lastPath.equals("") && order.equals("ALPHABETICAL") && path.toString().compareTo(lastPath) <= 0) {
                            return FileVisitResult.CONTINUE;
                        } else if (!lastPath.equals("") && order.equals("ASCEND_MODIFIED") && getLastModifiedTime(path.toString()).compareTo(getLastModifiedTime(lastPath)) <= 0) {
                            return FileVisitResult.CONTINUE;
                        } else if (!lastPath.equals("") && order.equals("DESCEND_MODIFIED") && getLastModifiedTime(path.toString()).compareTo(getLastModifiedTime(lastPath)) >= 0){
                            return FileVisitResult.CONTINUE;
                        } else if (!lastPath.equals("") && order.equals("ASCEND_CREATION") && getCreationTime(path.toString()).compareTo(getCreationTime(lastPath)) <= 0){
                            return FileVisitResult.CONTINUE;
                      } else if (!lastPath.equals("") && order.equals("DESCEND_MODIFIED") && getCreationTime(path.toString()).compareTo(getCreationTime(lastPath)) <= 0) {
                            return FileVisitResult.CONTINUE;
                        } else if (path.getFileName().toString().startsWith(".")) {
                             return FileVisitResult.CONTINUE;
                        } else {
                            if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                                  filesArray.add(path.toString());
                            }
                            return FileVisitResult.CONTINUE;
                        }
                    } catch ( IOException e){
                        throw new RuntimeException("IOException during the uploading files");
                    }
               }
            });
        } catch (IOException ex) {
            throw new RuntimeException(String.format("Failed get a list of local files at '%s'", directory), ex);
        }
        //return builder.build();
        return filesArray;
    }

    
    
    // Static method to return a FileTime of a file
    public static FileTime getCreationTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.creationTime();
      //System.out.println("The raw creation time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }
  
    public static FileTime getLastModifiedTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.lastModifiedTime();
      //System.out.println("The raw last modified time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }
}

interface Sorter {
    public static Comparator getComparator();
}

class AscendModifiedSorter implements Sorter {
    public static Comparator getComparator(){
        return new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
    }
}




class AscendCreationSorter implements Sorter {
    public static Comparator getComparator(){
        return new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getCreationTime(f1).compareTo(getCreationTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
    }
}

class DescendModifiedSorter implements Sorter {
    public static Comparator getComparator(){
        return new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return - getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
    }
}



class DescendCreationSorter implements Sorter {
    public static Comparator getComparator(){
        return new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return - getCreationTime(f1).compareTo(getCreationTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
    }
}